#include "filterintegrals.h"
#include "integrate.h"
#include "timer.h"

#include <algorithm>
#include <math.h>

using namespace std;

#ifdef FILTER_BOX

double lerp(double *v, double t)
{
	return (1-t)*v[0] + t*v[1];
}

double tv[2][2] = {{0,1},{1,0}};

double innerprod(double sa, double pa, double sb, double pb)
{
	double isa = 1. / sa;
	double isb = 1. / sb;

	double sum = 0;

	double ra0 = isa * (pa - .5);
	double ra1 = isa * (pa + .5);
	
	double rb0 = isb * (pb - .5);
	double rb1 = isb * (pb + .5);

	double r0 = max(ra0, rb0);
	double r1 = min(ra1, rb1);

	double r = r1 - r0;

	if (r > 0)
		sum += r;

	return sum;
}

double basis_times_basis(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)b(2^k*x-j-.5)
	double a2k = pow(2.,a[2]);
	double b2k = pow(2.,b[2]);

	return innerprod(a2k, -a[0]-.5, b2k, -b[0]-.5) *
		   innerprod(a2k, -a[1]-.5, b2k, -b[1]-.5);
}

double basis_times_filter(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)h(a*x+b)
	double a2k = pow(2.,a[2]);

	return innerprod(a2k, -a[0]-.5, b[2], -b[2]*b[0]) *
		   innerprod(a2k, -a[1]-.5, b[2], -b[2]*b[1]);
}

double filter_squared(double *a)
{
	// multiply functions with indices {i,j,k} -> h(a*x+b)h(a*x+b) == h(a*x)h(a*x)
	double v = innerprod(a[2], 0, a[2], 0);
	return v*v;
}

#elif defined FILTER_TENT

double lerp(double *v, double t)
{
	return (1-t)*v[0] + t*v[1];
}

double tv[2][2] = {{0,1},{1,0}};

double innerprod(double sa, double pa, double sb, double pb)
{
	double isa = 1. / sa;
	double isb = 1. / sb;

	double sum = 0;

	for (int i = 0; i < 2; i++)
	{
		for (int j = 0; j < 2; j++)
		{
			double ra0 = isa * (pa + i - 1);
			double ra1 = isa * (pa + i);
			
			double rb0 = isb * (pb + j - 1);
			double rb1 = isb * (pb + j);

			double r0 = max(ra0, rb0);
			double r1 = min(ra1, rb1);

			double r = r1 - r0;

			if (r > 0)
			{
				double ira = 1. / (ra1 - ra0);
				double irb = 1. / (rb1 - rb0);

				double ca0 = (r0 - ra0) * ira;
				double ca1 = (r1 - ra0) * ira;
				
				double cb0 = (r0 - rb0) * irb;
				double cb1 = (r1 - rb0) * irb;

				double va0 = lerp(tv[i], ca0);
				double va1 = lerp(tv[i], ca1);
				
				double vb0 = lerp(tv[j], cb0);
				double vb1 = lerp(tv[j], cb1);

				sum += r*(va0*(2*vb0 + vb1) + va1*(vb0 + 2*vb1))*(1./6.);
			}
		}
	}

	return sum;
}

double basis_times_basis(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)b(2^k*x-j-.5)
	double a2k = pow(2.,a[2]);
	double b2k = pow(2.,b[2]);

	return innerprod(a2k, -a[0]-.5, b2k, -b[0]-.5) *
		   innerprod(a2k, -a[1]-.5, b2k, -b[1]-.5);
}

double basis_times_filter(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)h(a*x+b)
	double a2k = pow(2.,a[2]);

	return innerprod(a2k, -a[0]-.5, b[2], -b[2]*b[0]) *
		   innerprod(a2k, -a[1]-.5, b[2], -b[2]*b[1]);
}

double filter_squared(double *a)
{
	// multiply functions with indices {i,j,k} -> h(a*x+b)h(a*x+b) == h(a*x)h(a*x)
	double v = innerprod(a[2], 0, a[2], 0);
	return v*v;
}

#elif defined FILTER_LANCZOS2 || defined FILTER_GAUSSIAN

static double pi = 3.14159265358979323846264338327950288;


#ifdef FILTER_LANCZOS2
static double l2(double x)
{
	if (x <= -2 || x >= 2)
		return 0;
	else if (fabs(x) < .0001)
		return 1 / 1.00979;
	else
		return 2*sin(pi*x/2) * sin(pi*x) / (pi*pi*x*x) / 1.00979;
}
#elif defined FILTER_GAUSSIAN
static double l2(double x)
{
	return exp(-2*x*x) * 0.797885;
}
#endif

struct FilterFunction
{
	double sa, pa, sb, pb;
	double operator()(double x)
	{
		return l2(sa*x + pa)*l2(sb*x + pb);
	}
};

double innerprod(double sa, double pa, double sb, double pb)
{
	const double r = 2;

	double minx = max((-r-pa)/sa, (-r-pb)/sb);
	double maxx = min((r-pa)/sa, (r-pb)/sb);

	// numerically integrate
	FilterFunction f;
	f.sa = sa;
	f.pa = pa;
	f.sb = sb;
	f.pb = pb;

	return integrate_trapezoid(f, minx, maxx, 100);
}

double basis_times_basis(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)b(2^k*x-j-.5)
	double a2k = pow(2.,a[2]);
	double b2k = pow(2.,b[2]);

	return innerprod(a2k, -a[0]-.5, b2k, -b[0]-.5) *
		   innerprod(a2k, -a[1]-.5, b2k, -b[1]-.5);
}

double basis_times_filter(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)h(a*x+b)
	double a2k = pow(2.,a[2]);

	return innerprod(a2k, -a[0]-.5, b[2], -b[2]*b[0]) *
		   innerprod(a2k, -a[1]-.5, b[2], -b[2]*b[1]);
}

double filter_squared(double *a)
{
	// multiply functions with indices {i,j,k} -> h(a*x+b)h(a*x+b) == h(a*x)h(a*x)
	double v = innerprod(a[2], 0, a[2], 0);
	return v*v;
}


#elif defined FILTER_JINC4 || defined FILTER_JINC3 || defined FILTER_JINC2

static double jinc1(double x)
{
	if (x < 1.2196698912665045)
		return -0.01010748302601642*(-1.219669891266506 + x)*(3.3469661812094422 + (-3.575439997501188 + x)*x)*
   (5.29736396423082 + (-3.3459258542770076 + x)*x)*(2.8530770806074077 + x*(1.366553807720093 + x))*
   (0.8017888983488939 + x*(1.6345956990078392 + x));
	else if (x < 2.2331305943815227)
		return 0.010731808029409387*(-2.233130594381783 + x)*(-1.2196698912665218 + x)*(-0.21939961378111664 + x)*
   (7.998642472168772 + (-5.606130181900144 + x)*x)*(9.395815619630644 + (-5.313454921619156 + x)*x)*
   (2.113806978804771 + (-0.8229263280834715 + x)*x);
	else
		return 0;
}

static double jinc2(double x)
{
	if (x < 1.2196698912665045)
		return -0.01010748302601642*(-1.219669891266506 + x)*(3.3469661812094422 + (-3.575439997501188 + x)*x)*
   (5.29736396423082 + (-3.3459258542770076 + x)*x)*(2.8530770806074077 + x*(1.366553807720093 + x))*
   (0.8017888983488939 + x*(1.6345956990078392 + x));
	else if (x < 2.2331305943815227)
		return 0.010731808029409387*(-2.233130594381783 + x)*(-1.2196698912665218 + x)*(-0.21939961378111664 + x)*
   (7.998642472168772 + (-5.606130181900144 + x)*x)*(9.395815619630644 + (-5.313454921619156 + x)*x)*
   (2.113806978804771 + (-0.8229263280834715 + x)*x);
	else
		return 0;
}

static double jinc3(double x)
{
	if (x < 1.2196698912665045)
		return -0.01010748302601642*(-1.219669891266506 + x)*(3.3469661812094422 + (-3.575439997501188 + x)*x)*
   (5.29736396423082 + (-3.3459258542770076 + x)*x)*(2.8530770806074077 + x*(1.366553807720093 + x))*
   (0.8017888983488939 + x*(1.6345956990078392 + x));
	else if (x < 2.2331305943815227)
		return 0.010731808029409387*(-2.233130594381783 + x)*(-1.2196698912665218 + x)*(-0.21939961378111664 + x)*
   (7.998642472168772 + (-5.606130181900144 + x)*x)*(9.395815619630644 + (-5.313454921619156 + x)*x)*
   (2.113806978804771 + (-0.8229263280834715 + x)*x);
	else if (x < 3.238315484166236)
		return -0.008515868104206192*(-3.238315484190778 + x)*(-2.233130594376959 + x)*(-1.5633793339029471 + x)*
   (14.981709707642281 + (-7.707716541350005 + x)*x)*(15.956187252744773 + (-7.412106246837379 + x)*x)*
   (3.769281429352575 + (-3.059082149138143 + x)*x);
	else
		return 0;
}

static double jinc4(double x)
{
	if (x < 1.2196698912665045)
		return -0.01010748302601642*(-1.219669891266506 + x)*(3.3469661812094422 + (-3.575439997501188 + x)*x)*
   (5.29736396423082 + (-3.3459258542770076 + x)*x)*(2.8530770806074077 + x*(1.366553807720093 + x))*
   (0.8017888983488939 + x*(1.6345956990078392 + x));
	else if (x < 2.2331305943815227)
		return 0.010731808029409387*(-2.233130594381783 + x)*(-1.2196698912665218 + x)*(-0.21939961378111664 + x)*
   (7.998642472168772 + (-5.606130181900144 + x)*x)*(9.395815619630644 + (-5.313454921619156 + x)*x)*
   (2.113806978804771 + (-0.8229263280834715 + x)*x);
	else if (x < 3.238315484166236)
		return -0.008515868104206192*(-3.238315484190778 + x)*(-2.233130594376959 + x)*(-1.5633793339029471 + x)*
   (14.981709707642281 + (-7.707716541350005 + x)*x)*(15.956187252744773 + (-7.412106246837379 + x)*x)*
   (3.769281429352575 + (-3.059082149138143 + x)*x);
	else if (x < 4.24106286379607)
		return 0.005949486328726622*(-4.241062862825876 + x)*(-3.238315484393466 + x)*(-2.6539598424833133 + x)*
   (24.992832757974156 + (-9.983172253359422 + x)*x)*(24.941616261506358 + (-9.595399297405045 + x)*x)*
   (8.231230303083018 + (-5.287739288329403 + x)*x);
	else
		return 0;
}

static double filter(double x, double y)
{
	double d = sqrt(x*x + y*y);
#ifdef FILTER_JINC4
	return jinc4(d) * jinc1(d * 1.2196698912665045 / 4.24106286379607) / 0.319636;
#elif defined FILTER_JINC3
	return jinc3(d) * jinc1(d * 1.2196698912665045 / 3.238315484166236) / 0.315776;
#else
	return jinc2(d) * jinc1(d * 1.2196698912665045 / 2.2331305943815227) / 0.325394;
#endif
}

struct FilterFunction
{
	double sa, pax, pay, sb, pbx, pby;
	double operator()(double x, double y)
	{
		double ax = sa*x + pax;
		double ay = sa*y + pay;
		//double ad = sqrt(ax*ax + ay*ay);

		double bx = sb*x + pbx;
		double by = sb*y + pby;
		//double bd = sqrt(bx*bx + by*by);

		//return jinc3(ad)*jinc3(bd);
		return filter(ax, ay)*filter(bx, by);
	}
};

double innerprod(double sa, double pax, double pay, double sb, double pbx, double pby)
{
#ifdef FILTER_JINC4
	const double r = 4.24106286379607;
#elif defined FILTER_JINC3
	const double r = 3.238315484166236;
#else
	const double r = 2.2331305943815227;
#endif

	double minx = max((-r-pax)/sa, (-r-pbx)/sb);
	double maxx = min((r-pax)/sa, (r-pbx)/sb);
	
	double miny = max((-r-pay)/sa, (-r-pby)/sb);
	double maxy = min((r-pay)/sa, (r-pby)/sb);

	// numerically integrate
	FilterFunction f;
	f.sa = sa;
	f.pax = pax;
	f.pay = pay;
	f.sb = sb;
	f.pbx = pbx;
	f.pby = pby;

	return integrate2_box(f, minx, maxx, miny, maxy, 50, 50);
}

double basis_times_basis(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)b(2^k*x-j-.5)
	double a2k = pow(2.,a[2]);
	double b2k = pow(2.,b[2]);

	return innerprod(a2k, -a[0]-.5, -a[1]-.5, b2k, -b[0]-.5, -b[1]-.5);
}

double basis_times_filter(double *a, double *b)
{
	// multiply functions with indices {i,j,k} -> b(2^k*x-i-.5)h(a*x+b)
	double a2k = pow(2.,a[2]);

	return innerprod(a2k, -a[0]-.5, -a[1]-.5, b[2], -b[2]*b[0], -b[2]*b[1]);
}

double filter_squared(double *a)
{
	// multiply functions with indices {i,j,k} -> h(a*x+b)h(a*x+b) == h(a*x)h(a*x)
	return innerprod(a[2], 0, 0, a[2], 0, 0);
}

#endif
