#include "optimize.h"

#include <stdio.h>

GridBasisIndices grid_basis_indices[5];

double frand0()
{
	return 2.0*rand() / (double)(RAND_MAX) - 1.0;
}

void choose_basis_functions(Array<BasisParam> &basisparams)
{
	// these basis functions will be shared among all calculations
	basisparams.s = 0;
	BasisParam b;

	for (b.k = 0; b.k <= max_mip_level; b.k++)
	{
		int minidx, maxidx;

		if (b.k == 0)
		{
			minidx = -1;
			maxidx = 0;
		}
		else if (b.k == 1)
		{
			minidx = -2;
			maxidx = 2;
		}
		else if (b.k == 2)
		{
			minidx = -4;
			maxidx = 5;
		}

		grid_basis_indices[b.k].minidx = minidx;
		grid_basis_indices[b.k].maxidx = maxidx;
		grid_basis_indices[b.k].grid.resize(maxidx - minidx + 1, maxidx - minidx + 1);
		
		for (b.i = minidx; b.i <= maxidx; b.i++)
		{
			for (b.j = minidx; b.j <= maxidx; b.j++)
			{
				grid_basis_indices[b.k].grid(b.i-minidx, b.j-minidx) = basisparams.size();
				basisparams.push_back(b);
			}
		}
	}

	printf("There are %d available basis functions.\n", basisparams.s);
}


void choose_hw_basis_functions(Array<BasisParam> &basisparams)
{
	// these basis functions will be shared among all calculations
	basisparams.s = 0;
	BasisParam b;

	for (b.k = 0; b.k <= max_mip_level; b.k++)
	{
		int minidx, maxidx;

		const int boundary = 1;
		if (b.k == 0)
		{
			minidx = -1 - boundary;
			maxidx = 0 + boundary;
		}
		else if (b.k == 1)
		{
			minidx = -2 - boundary;
			maxidx = 2 + boundary;
		}
		else if (b.k == 2)
		{
			minidx = -4 - boundary;
			maxidx = 5 + boundary;
		}

		grid_basis_indices[b.k].minidx = minidx;
		grid_basis_indices[b.k].maxidx = maxidx;
		grid_basis_indices[b.k].grid.resize(maxidx - minidx + 1, maxidx - minidx + 1);
		
		for (b.i = minidx; b.i <= maxidx; b.i++)
		{
			for (b.j = minidx; b.j <= maxidx; b.j++)
			{
				grid_basis_indices[b.k].grid(b.i-minidx, b.j-minidx) = basisparams.size();
				basisparams.push_back(b);
			}
		}
	}

	printf("There are %d available basis functions.\n", basisparams.s);
}


int find_idx(Array<BasisParam> &basisparams, BasisParam &b)
{
	for (int j = 0; j < basisparams.s; j++)
	{
		if (basisparams[j] == b)
			return j;
	}
	assert(false);
	return -1;
}

void choose_bilinear_basis_indices(Array<BasisIndexBi> &basisindexbi, Array<BasisParam> &basisparams)
{
	// first create params for the bottom left of the bilinear samples
	Array<BasisParam> bps;
	{
		BasisParam b;

		for (b.k = 0; b.k <= max_mip_level; b.k++)
		{
			int minidx, maxidx;

			if (b.k == 0)
			{
				minidx = -1;
				maxidx = 0 - 1;
			}
			else if (b.k == 1)
			{
				minidx = -2;
				maxidx = 2 - 1;
			}
			else if (b.k == 2)
			{
				minidx = -4;
				maxidx = 5 - 1;
			}
			
			for (b.i = minidx; b.i <= maxidx; b.i++)
			{
				for (b.j = minidx; b.j <= maxidx; b.j++)
				{
					bps.push_back(b);
				}
			}
		}
	}

	// now find the indices of the bilinear functions
	basisindexbi.s = 0;
	for (int i = 0; i < bps.s; i++)
	{
		BasisIndexBi &bi = basisindexbi.push_back();

		BasisParam b = bps[i];
		bi.idx[0] = find_idx(basisparams, b);

		b.i++;
		bi.idx[1] = find_idx(basisparams, b);

		b.j++;
		bi.idx[3] = find_idx(basisparams, b);

		b.i--;
		bi.idx[2] = find_idx(basisparams, b);
	}

	printf("There are %d available bilinear functions.\n", basisindexbi.s);
}

void build_overlaps(Array3D<Sample> &samples, Array2D<double> &bb, Array<BasisParam> &basisparams, double poi, double psi, double poj, double psj, double pok, double psk, int ns)
{
	// build the basis X basis matrix
	double a[3], b[3], h[3], t[3];
	bb.resize(basisparams.s, basisparams.s);

	for (int bi = 0; bi < basisparams.s; bi++)
	{
		a[0] = basisparams[bi].i;
		a[1] = basisparams[bi].j;
		a[2] = basisparams[bi].k;

		for (int bj = 0; bj < basisparams.s; bj++)
		{
			b[0] = basisparams[bj].i;
			b[1] = basisparams[bj].j;
			b[2] = basisparams[bj].k;

			bb(bi,bj) = basis_times_basis(a,b) * (pow(4.0, basisparams[bi].k) * pow(4.0, basisparams[bj].k));
		}
	}

	// build h dependent variables
	samples.resize(ns, ns, ns);

	for (int sk = 0; sk < ns; sk++)
	{
		t[2] = (sk+.5) / ns;
		h[2] = pow(2.0, pok + psk * t[2]);

		for (int sj = 0; sj < ns; sj++)
		{
			t[1] = (sj+.5) / ns;
			h[1] = poj + psj * t[1];

			for (int si = 0; si < ns; si++)
			{
				t[0] = (si+.5) / ns;
				h[0] = poi + psi * t[0];

				Sample &s = samples(si, sj, sk);
							
				s.t[0] = t[0];
				s.t[1] = t[1];
				s.t[2] = t[2];

				s.h[0] = h[0];
				s.h[1] = h[1];
				s.h[2] = h[2];

				s.hb.resize(basisparams.s);

				s.hh = filter_squared(h) * (h[2]*h[2]*h[2]*h[2]);

				for (int bi = 0; bi < basisparams.s; bi++)
				{
					a[0] = basisparams[bi].i;
					a[1] = basisparams[bi].j;
					a[2] = basisparams[bi].k;

					s.hb[bi] = basis_times_filter(a,h) * (pow(4.0, basisparams[bi].k) * h[2]*h[2]);
				}
			}
		}
	}
}



void vec_print(double *A, int s, FILE *fa)
{
	// print matrix
	fprintf(fa, "{");
	for (int j = 0; j < s; j++)
	{
		if (j != 0)
			fprintf(fa, ",");
		fprintf(fa, "%16.15f", A[j]);
	}
	fprintf(fa, "}");
}

void vec_print(double *A, int s, string fn)
{
	// print matrix
	FILE *fa = fopen(fn.c_str(), "wb");
	vec_print(A, s, fa);
	fclose(fa);
}

void vec_print(Array<double> &A, string fn)
{
	// print matrix
	vec_print(A.data, A.s, fn);
}

void mat_print(Array2D<double> &A, string fn)
{
	// print matrix
	FILE *fa = fopen(fn.c_str(), "wb");
	fprintf(fa, "{");
	for (int i = 0; i < A.s[0]; i++)
	{
		if (i != 0)
			fprintf(fa, ",{");
		else
			fprintf(fa, "{");

		for (int j = 0; j < A.s[1]; j++)
		{
			if (j != 0)
				fprintf(fa, ",");
			fprintf(fa, "%16.15f", A(i,j));
		}
		
		fprintf(fa, "}");
	}
	fprintf(fa, "}");
	fclose(fa);
}
